@@ -2,7 +2,7 @@
  *
  *  Connection Manager
  *
- *  Copyright (C) 2007-2009  Intel Corporation. All rights reserved.
+ *  Copyright (C) 2007-2010  Intel Corporation. All rights reserved.
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License version 2 as
@@ -43,11 +43,6 @@ enum connman_device_type {
 	CONNMAN_DEVICE_TYPE_BLUETOOTH = 4,
 	CONNMAN_DEVICE_TYPE_CELLULAR  = 5,
 	CONNMAN_DEVICE_TYPE_GPS       = 6,
-	CONNMAN_DEVICE_TYPE_MBM       = 14,
-	CONNMAN_DEVICE_TYPE_HSO       = 15,
-	CONNMAN_DEVICE_TYPE_NOZOMI    = 16,
-	CONNMAN_DEVICE_TYPE_HUAWEI    = 17,
-	CONNMAN_DEVICE_TYPE_NOVATEL   = 18,
 	CONNMAN_DEVICE_TYPE_VENDOR    = 10000,
 };
 
@@ -55,9 +50,12 @@ enum connman_device_mode {
 	CONNMAN_DEVICE_MODE_UNKNOWN          = 0,
 	CONNMAN_DEVICE_MODE_NETWORK_SINGLE   = 1,
 	CONNMAN_DEVICE_MODE_NETWORK_MULTIPLE = 2,
-	CONNMAN_DEVICE_MODE_TRANSPORT_IP     = 3,
 };
 
+#define CONNMAN_DEVICE_PRIORITY_LOW      -100
+#define CONNMAN_DEVICE_PRIORITY_DEFAULT     0
+#define CONNMAN_DEVICE_PRIORITY_HIGH      100
+
 struct connman_device;
 
 struct connman_device *connman_device_create(const char *node,
@@ -76,21 +74,16 @@ const char *connman_device_get_control(struct connman_device *device);
 
 void connman_device_set_ident(struct connman_device *device,
 						const char *ident);
+const char *connman_device_get_ident(struct connman_device *device);
 
 void connman_device_set_mode(struct connman_device *device,
 					enum connman_device_mode mode);
 enum connman_device_mode connman_device_get_mode(struct connman_device *device);
-void connman_device_set_secondary(struct connman_device *device,
-						connman_bool_t secondary);
-connman_bool_t connman_device_get_secondary(struct connman_device *device);
 
 int connman_device_set_powered(struct connman_device *device,
 						connman_bool_t powered);
-int connman_device_set_carrier(struct connman_device *device,
-						connman_bool_t carrier);
 int connman_device_set_scanning(struct connman_device *device,
 						connman_bool_t scanning);
-void connman_device_cleanup_scanning(struct connman_device *device);
 
 int connman_device_set_disconnected(struct connman_device *device,
 						connman_bool_t disconnected);
@@ -127,11 +120,7 @@ struct connman_device_driver {
 	void (*remove) (struct connman_device *device);
 	int (*enable) (struct connman_device *device);
 	int (*disable) (struct connman_device *device);
-	int (*connect) (struct connman_device *device);
-	int (*disconnect) (struct connman_device *device);
 	int (*scan) (struct connman_device *device);
-	int (*join) (struct connman_device *device,
-					struct connman_network *network);
 };
 
 int connman_device_driver_register(struct connman_device_driver *driver);
