@@ -33,6 +33,7 @@
 struct gateway_data {
 	int index;
 	char *ipv4_gateway;
+	char *ipv6_gateway;
 	struct connman_element *element;
 	unsigned int order;
 	gboolean active;
@@ -77,6 +78,10 @@ static int del_routes(struct gateway_data *data)
 	} else if (g_strcmp0(data->ipv4_gateway, "0.0.0.0") == 0) {
 		return connman_inet_clear_gateway_interface(data->index);
 	} else {
+		connman_inet_del_ipv6_host_route(data->index,
+						data->ipv6_gateway);
+		connman_inet_clear_ipv6_gateway_address(data->index,
+							data->ipv6_gateway);
 		connman_inet_del_host_route(data->index, data->ipv4_gateway);
 		return connman_inet_clear_gateway_address(data->index,
 							data->ipv4_gateway);
@@ -98,11 +103,15 @@ static void find_element(struct connman_element *element, gpointer user_data)
 	data->element = element;
 }
 
-static struct gateway_data *add_gateway(int index, const char *gateway)
+static struct gateway_data *add_gateway(int index, const char *gateway,
+						const char *ipv6_gateway)
 {
 	struct gateway_data *data;
 	struct connman_service *service;
 
+	DBG("index %d ipv4 gateway %s ipv6 gateway %s", index, gateway,
+							ipv6_gateway);
+
 	if (strlen(gateway) == 0)
 		return NULL;
 
@@ -112,6 +121,7 @@ static struct gateway_data *add_gateway(int index, const char *gateway)
 
 	data->index = index;
 	data->ipv4_gateway = g_strdup(gateway);
+	data->ipv6_gateway = g_strdup(ipv6_gateway);
 	data->active = FALSE;
 	data->element = NULL;
 	data->vpn_ip = NULL;
@@ -152,7 +162,13 @@ static void set_default_gateway(struct gateway_data *data)
 	if (data->vpn == TRUE) {
 		connman_inet_set_gateway_address(data->index, data->vpn_ip);
 		data->active = TRUE;
-		/* vpn gateway going away no changes in services */
+
+		service = __connman_service_lookup_from_index(data->index);
+		if (service == NULL)
+			return;
+
+		__connman_service_indicate_default(service);
+
 		return;
 	}
 
@@ -162,6 +178,8 @@ static void set_default_gateway(struct gateway_data *data)
 		goto done;
 	}
 
+	connman_inet_set_ipv6_gateway_address(element->index,
+						data->ipv6_gateway);
 	if (connman_inet_set_gateway_address(element->index,
 					data->ipv4_gateway) < 0)
 		return;
@@ -249,8 +267,9 @@ static struct gateway_data *find_active_gateway(void)
 static int connection_probe(struct connman_element *element)
 {
 	struct connman_service *service = NULL;
-	const char *gateway = NULL;
+	const char *gateway = NULL, *ipv6_gateway = NULL;
 	const char *vpn_ip = NULL;
+	const char *domainname = NULL;
 	struct gateway_data *active_gateway = NULL;
 	struct gateway_data *new_gateway = NULL;
 
@@ -265,11 +284,23 @@ static int connection_probe(struct connman_element *element)
 
 	connman_element_get_value(element,
 				CONNMAN_PROPERTY_ID_IPV4_GATEWAY, &gateway);
+	connman_element_get_value(element,
+			CONNMAN_PROPERTY_ID_IPV6_GATEWAY, &ipv6_gateway);
 
 	connman_element_get_value(element,
-				  CONNMAN_PROPERTY_ID_IPV4_ADDRESS, &vpn_ip);
+			CONNMAN_PROPERTY_ID_IPV4_PEER, &vpn_ip);
 
-	DBG("gateway %s", gateway);
+	if (vpn_ip == NULL)
+		connman_element_get_value(element,
+				CONNMAN_PROPERTY_ID_IPV4_ADDRESS, &vpn_ip);
+
+	DBG("vpn_ip %s", vpn_ip);
+
+	connman_element_get_value(element,
+			CONNMAN_PROPERTY_ID_DOMAINNAME, &domainname);
+
+	DBG("ipv4 gateway %s ipv6 gateway %s domainname %s",
+				gateway, ipv6_gateway, domainname);
 
 	/*
 	 * If gateway is NULL, it's a point to point link and the default
@@ -283,16 +314,19 @@ static int connection_probe(struct connman_element *element)
 	connman_element_set_enabled(element, TRUE);
 
 	active_gateway = find_active_gateway();
-	new_gateway = add_gateway(element->index, gateway);
+	new_gateway = add_gateway(element->index, gateway, ipv6_gateway);
 	if (new_gateway == NULL)
 		return 0;
 
 	service = __connman_element_get_service(element);
 
+	connman_inet_add_ipv6_host_route(element->index,
+					new_gateway->ipv6_gateway, NULL);
 	connman_inet_add_host_route(element->index,
 					new_gateway->ipv4_gateway, NULL);
 	__connman_service_nameserver_add_routes(service,
 						new_gateway->ipv4_gateway);
+	__connman_service_set_domainname(service, domainname);
 
 	__connman_service_indicate_state(service, CONNMAN_SERVICE_STATE_READY);
 
@@ -422,12 +456,13 @@ static void update_order(void)
 	for (list = gateway_list; list; list = list->next) {
 		struct gateway_data *data = list->data;
 		struct connman_service *service;
+		int index = data->index;
 
-		/* vpn gataway is not attached to a service. */
 		if (data->vpn)
-			continue;
+			service = __connman_service_lookup_from_index(index);
+		else
+			service = __connman_element_get_service(data->element);
 
-		service = __connman_element_get_service(data->element);
 		data->order = __connman_service_get_order(service);
 	}
 }
